/*
 * File:   weight_acts_kernel.cu
 * Author: Alex Krizhevsky
 *
 * Created on February 2, 2011, 4:43 PM
 */

#include "../../include/kernel/weight_acts_kernel.cuh"

/*
 * Each block computes weight gradients for B_Y * pixelsPerThread pixels and B_X filters
 * threadIdx.x determines filter
 * threadIdx.y determines pixel in filter
 *
 * blockIdx.x determines filter batch of B_X
 * blockIdx.y determines pixel batch of B_Y * pixelsPerThread
 *
 * Number of filters must be divisible by B_X
 * Number of images (cases) should be divisible by preloadCases.
 *
 * images:      (numColors, imgPixels, numImages), with stride given
 * hidActs:     (numModules, numFilters, numImages) if mfi
 *              (numFilters, numModules, numImages) otherwise
 *
 * targets:     (numModules, numColors, filterPixels, numFilters)
 *
 * B_Y * B_X should be divisible by preloadCases.
 * preloadCases one of 16, 32.
 * B_X one of 4, 8, 16, 32
 * B_Y arbitrary (satisfying divisibility constraints)
 *
 * After adding pixelsPerThread, register usage went from 20 to 23 (when pixelsPerThread = 1)...
 * so the compiler is messing up here somehow. It's unable to optimize that case away.
 */
template <int B_Y, int B_X, int pixelsPerThread, int preloadCases, int numColors, bool mfi>
__global__ void weight_acts_kernel_color(float* images, float* hidActs, float* targets,
                                         const int numImages, const int numFilters,
                                         const int numModulesX,
                                         const int imgSize, const int filterSize,
                                         const int paddingStart, const int moduleStride, const int imgStride) {
    __shared__ float shImgGrads[pixelsPerThread * B_Y * numColors][preloadCases]; // preload preloadCases cases of B_Y * pixelsPerThread pixels
    __shared__ float shHid2Acts[B_X][preloadCases + 1]; // preload preloadCases cases of B_X hid2 acts

    const int tidx = B_X * threadIdx.y + threadIdx.x;
    const int loadY = tidx / preloadCases, loadX = tidx % preloadCases;

    const int filterPixels = filterSize * filterSize;
    const int imgPixels = imgSize * imgSize;

    const int blocksPerModule = numFilters / B_X;
    const int moduleIdx = blockIdx.x / blocksPerModule;
    const int blockFilterIdx = B_X * (blockIdx.x % blocksPerModule);

//    const int moduleStride = (imgSize - filterSize + 1) / numModulesX; 
//    const int numModules = MUL24(numModulesX, numModulesX);

    int imgLoadModPosY = paddingStart + (moduleIdx / numModulesX) * moduleStride;
    int imgLoadModPosX = paddingStart + (moduleIdx % numModulesX) * moduleStride;
    const int blockPixelOffset = blockIdx.y * B_Y * pixelsPerThread;

    images += loadX;
    if (mfi) {
        hidActs += moduleIdx * numFilters * numImages
                + blockFilterIdx * numImages
                + loadY * numImages
                + loadX;
    } else {
        hidActs += moduleIdx * numImages
                + blockFilterIdx * numImages * numModulesX * numModulesX
                + loadY * numImages * numModulesX * numModulesX
                + loadX;
    }
    targets += (moduleIdx * numFilters) * filterPixels * numColors
            + blockPixelOffset * numFilters
            + blockFilterIdx
            + threadIdx.y * numFilters + threadIdx.x;

    float* shImgGradLoad = &shImgGrads[loadY][loadX];
    float* shHidActLoad = &shHid2Acts[loadY][loadX];

    float prod[numColors][pixelsPerThread];
    #pragma unroll
    for (int c = 0; c < numColors; c++) {
        #pragma unroll
        for (int p = 0; p < pixelsPerThread; p++) {
            prod[c][p] = 0;
        }
    }
    for (int caseIdx = 0; caseIdx < numImages; caseIdx += preloadCases) {
        if (loadY < B_Y * pixelsPerThread) {
            /*
             * As long as B_Y * B_X is divisible by preloadCases this will loop the right
             * number of times.
             *
             * This will load some imgGrads from filter pixels that don't exit (it'll set those to 0),
             * but the code does not produce any output for those pixels (see last lines).
             */
//            #pragma unroll
            for (int y = 0; y < B_Y * pixelsPerThread; y += (B_X * B_Y) / preloadCases) {
                // Make sure number of rows in the array is divisible by number of rows filled per iteration
                if ((B_Y * pixelsPerThread) % (B_X * B_Y / preloadCases) == 0 || y + loadY < B_Y * pixelsPerThread) {
                    const int pxIdx = blockPixelOffset + loadY + y; // pixel idx in filter

                    if (pxIdx < filterPixels) {
                        const int pxY = imgLoadModPosY + pxIdx / filterSize; // pixel x,y coords in image
                        const int pxX = imgLoadModPosX + pxIdx % filterSize;
                        if (pxY >= 0 && pxY < imgSize && pxX >= 0 && pxX < imgSize) {
                            const int pixIdx = (pxY * imgSize + pxX) * imgStride;
                            #pragma unroll
                            for(int c = 0; c < numColors; c++) {
                                shImgGradLoad[(y + c * pixelsPerThread * B_Y) * preloadCases] = images[c * imgPixels * imgStride + pixIdx];
                            }
                        } else {
                            #pragma unroll
                            for(int c = 0; c < numColors; c++) {
                                shImgGradLoad[(y + c * pixelsPerThread * B_Y) * preloadCases] = 0;
                            }
                        }
                    } else {
                        #pragma unroll
                        for(int c = 0; c < numColors; c++) {
                            shImgGradLoad[(y + c * pixelsPerThread * B_Y) * preloadCases] = 0;
                        }
                    }
                }
            }
        }
        if (loadY < B_X) {
            if (mfi) {
                #pragma unroll
                for (int y = 0; y < B_X; y += (B_X * B_Y) / preloadCases) {
                    // Make sure number of rows in the array is divisible by number of rows filled per iteration
                    if (B_X % (B_X * B_Y / preloadCases) == 0 || y + loadY < B_X) {
                        shHidActLoad[y * (preloadCases + 1)] = hidActs[y * numImages];
                    }
                }
            } else {
                #pragma unroll
                for (int y = 0; y < B_X; y += (B_X * B_Y) / preloadCases) {
                    // Make sure number of rows in the array is divisible by number of rows filled per iteration
                    if (B_X % (B_X * B_Y / preloadCases) == 0 || y + loadY < B_X) {
                        shHidActLoad[y * (preloadCases + 1)] = hidActs[y * numImages * numModulesX * numModulesX];
                    }
                }
            }
            hidActs += preloadCases;
        }

        __syncthreads();
        #pragma unroll
        for (int c = 0; c < numColors; c++) {
            #pragma unroll
            for (int p = 0; p < pixelsPerThread; p++) {
                #pragma unroll
                for (int i = 0; i < preloadCases; i++) {
                    prod[c][p] += shImgGrads[threadIdx.y + p * B_Y + c * pixelsPerThread * B_Y][i] * shHid2Acts[threadIdx.x][i];
                }
            }
        }
        __syncthreads();
        images += preloadCases;
    }
    
    #pragma unroll
    for (int p = 0; p < pixelsPerThread; p++) {
        if (blockPixelOffset + p * B_Y + threadIdx.y < filterPixels) {
            #pragma unroll
            for (int c = 0; c < numColors; c++) {
                targets[p * B_Y * numFilters + c * filterPixels * numFilters] = prod[c][p];
            }
        }
    }
}

/*
 * Each block computes weight gradients for B_Y * pixelsPerThread pixels and B_X filters
 * threadIdx.x determines filter
 * threadIdx.y determines pixel in filter
 *
 * blockIdx.x determines filter batch of B_X
 * blockIdx.y determines pixel, color batch of B_Y * pixelsPerThread * colorsPerThread
 *      In essence, blockIdx.y.x = 0...numColors / colorsPerThread
 *                  blockIdx.y.y = 0...DIVUP(numPixels, B_Y*pixelsPerThread)
 *
 * Number of filters must be divisible by B_X
 * Number of images (cases) should be divisible by preloadCases.
 *
 * images:      (numColors, imgPixels, numImages), with stride given
 * hidActs:     (numModules, numFilters, numImages) if mfi
 *              (numFilters, numModules, numImages) otherwise
 *
 * targets:     (numModules, numColors, filterPixels, numFilters)
 *
 * B_Y * B_X should be divisible by preloadCases.
 * preloadCases one of 16, 32.
 * B_X one of 4, 8, 16, 32
 * B_Y arbitrary (satisfying divisibility constraints)
 * Number of colors should be divisible by colorsPerThread.
 *
 * After adding pixelsPerThread, register usage went from 20 to 23 (when pixelsPerThread = 1)...
 * so the compiler is messing up here somehow. It's unable to optimize that case away.
 */
template <int B_Y, int B_X, int pixelsPerThread, int colorsPerThread, int preloadCases, bool mfi>
__global__ void weight_acts_kernel_manycolor(float* images, float* hidActs, float* targets,
                                         const int numImages, const int numFilters,
                                         const int numModulesX,
                                         const int imgSize, const int filterSize,
                                         const int paddingStart, const int moduleStride, const int imgStride,
                                         const int numColors) {
    __shared__ float shImgGrads[colorsPerThread * pixelsPerThread * B_Y][preloadCases]; // preload preloadCases cases of B_Y * pixelsPerThread pixels
    __shared__ float shHid2Acts[B_X][preloadCases + 1]; // preload preloadCases cases of B_X hid2 acts

    const int tidx = B_X * threadIdx.y + threadIdx.x;
    const int loadY = tidx / preloadCases, loadX = tidx % preloadCases;

    const int filterPixels = filterSize * filterSize;
    const int imgPixels = imgSize * imgSize;

    const int blocksPerModule = numFilters / B_X;
    const int moduleIdx = blockIdx.x / blocksPerModule;
    const int blockFilterIdx = B_X * (blockIdx.x % blocksPerModule);

    int imgLoadModPosY = paddingStart + (moduleIdx / numModulesX) * moduleStride;
    int imgLoadModPosX = paddingStart + (moduleIdx % numModulesX) * moduleStride;
    const int blockPixelOffset = (blockIdx.y / (numColors/colorsPerThread)) * B_Y * pixelsPerThread;
    const int blockColorOffset = (blockIdx.y % (numColors/colorsPerThread)) * colorsPerThread;

    images += blockColorOffset * imgPixels * imgStride + loadX;
    if (mfi) {
        hidActs += moduleIdx * numFilters * numImages
                + blockFilterIdx * numImages
                + loadY * numImages
                + loadX;
    } else {
        hidActs += moduleIdx * numImages
                + blockFilterIdx * numImages * numModulesX * numModulesX
                + loadY * numImages * numModulesX * numModulesX
                + loadX;
    }
    targets += moduleIdx * numFilters * filterPixels * numColors
            + blockColorOffset * filterPixels * numFilters
            + blockPixelOffset * numFilters
            + blockFilterIdx
            + threadIdx.y * numFilters + threadIdx.x;

    float* shImgGradLoad = &shImgGrads[loadY][loadX];
    float* shHidActLoad = &shHid2Acts[loadY][loadX];

    float prod[colorsPerThread][pixelsPerThread];
    #pragma unroll
    for (int c = 0; c < colorsPerThread; c++) {
        #pragma unroll
        for (int p = 0; p < pixelsPerThread; p++) {
            prod[c][p] = 0;
        }
    }
    for (int caseIdx = 0; caseIdx < numImages; caseIdx += preloadCases) {
        if (loadY < B_Y * pixelsPerThread) {
            /*
             * As long as B_Y * B_X is divisible by preloadCases this will loop the right
             * number of times.
             *
             * This will load some imgGrads from filter pixels that don't exit (it'll set those to 0),
             * but the code does not produce any output for those pixels (see last lines).
             */
//            #pragma unroll
            for (int y = 0; y < B_Y * pixelsPerThread; y += (B_X * B_Y) / preloadCases) {
                // Make sure number of rows in the array is divisible by number of rows filled per iteration
                if ((B_Y * pixelsPerThread) % (B_X * B_Y / preloadCases) == 0 || y + loadY < B_Y * pixelsPerThread) {
                    const int pxIdx = blockPixelOffset + loadY + y; // pixel idx in filter

                    if (pxIdx < filterPixels) {
                        const int pxY = imgLoadModPosY + pxIdx / filterSize; // pixel x,y coords in image
                        const int pxX = imgLoadModPosX + pxIdx % filterSize;
                        if (pxY >= 0 && pxY < imgSize && pxX >= 0 && pxX < imgSize) {
                            const int pixIdx = (pxY * imgSize + pxX) * imgStride; // pixel idx in image
                            #pragma unroll
                            for (int c = 0; c < colorsPerThread; c++) {
                                shImgGradLoad[(y + c * pixelsPerThread * B_Y) * preloadCases] = images[c * imgPixels * imgStride + pixIdx];
                            }
                        } else {
                            #pragma unroll
                            for (int c = 0; c < colorsPerThread; c++) {
                                shImgGradLoad[(y + c * pixelsPerThread * B_Y) * preloadCases] = 0;
                            }
                        }
                    } else {
                        #pragma unroll
                        for (int c = 0; c < colorsPerThread; c++) {
                            shImgGradLoad[(y + c * pixelsPerThread * B_Y) * preloadCases] = 0;
                        }
                    }
                }
            }
        }
        if (loadY < B_X) {
            if (mfi) {
                #pragma unroll
                for (int y = 0; y < B_X; y += (B_X * B_Y) / preloadCases) {
                    // Make sure number of rows in the array is divisible by number of rows filled per iteration
                    if (B_X % (B_X * B_Y / preloadCases) == 0 || y + loadY < B_X) {
                        shHidActLoad[y * (preloadCases + 1)] = hidActs[y * numImages];
                    }
                }
            } else {
                #pragma unroll
                for (int y = 0; y < B_X; y += (B_X * B_Y) / preloadCases) {
                    // Make sure number of rows in the array is divisible by number of rows filled per iteration
                    if (B_X % (B_X * B_Y / preloadCases) == 0 || y + loadY < B_X) {
                        shHidActLoad[y * (preloadCases + 1)] = hidActs[y * numImages * numModulesX * numModulesX];
                    }
                }
            }
            hidActs += preloadCases;
        }

        __syncthreads();

        #pragma unroll
        for (int c = 0; c < colorsPerThread; c++) {
            #pragma unroll
            for (int p = 0; p < pixelsPerThread; p++) {
                #pragma unroll
                for (int i = 0; i < preloadCases; i++) {
                    prod[c][p] += shImgGrads[threadIdx.y + p * B_Y + c * pixelsPerThread * B_Y][i] * shHid2Acts[threadIdx.x][i];
                }
            }
        }
        __syncthreads();
        images += preloadCases;
    }

    #pragma unroll
    for (int p = 0; p < pixelsPerThread; p++) {
        if (blockPixelOffset + p * B_Y + threadIdx.y < filterPixels) {
            #pragma unroll
            for (int c = 0; c < colorsPerThread; c++) {
                targets[p * B_Y * numFilters + c * filterPixels * numFilters] = prod[c][p];
            }
        }
    }
}

/*
 * images:      (numColors, imgPixels, numImages), with stride given
 * hidActs:     (numModules, numFilters, numImages) if hidActsOrder == MODULE_FILTER_IMAGE
 *              (numFilters, numModules, numImages) otherwise
 *
 * targets:     (numModules, numColors, filterPixels, numFilters)
 *
 * images: The images matrix.
 * hidActs: The filter activity matrix.
 * targets: Result matrix.
 * numModulesX: number of filter applications in the x (or equivalently y) dimension. So the total
 *              number of modules will be the square of this number.
 * filterSize: the width (or equivalently height) of the filter.
 * paddingStart: non-positive number indicating where the first filter should be applied.
 * moduleStride: stride between filter applications.
 * numColors: number of color channels in images and filters.
 * hidActsOrder: how the hidActs matrix is laid out (see hidActs comment above).
 */
void computeWeightActs(NVMatrix& images, NVMatrix& hidActs, NVMatrix& targets,
                       int numModulesX, int filterSize, int paddingStart, int moduleStride, int numColors, FILTER_OUTPUT_ORDER hidActsOrder) {
    assert(numColors > 0 && (numColors <= 3 || numColors % 4 == 0));

    assert(paddingStart <= 0);

    int imgStride = images.getStride();
    int numImages = images.getNumCols();
    int imgPixels = images.getNumRows() / numColors;
    int imgSize = int(sqrt(imgPixels));
    int numModules = numModulesX * numModulesX;
    assert(imgSize * imgSize == imgPixels);
    assert(images.getNumRows() == imgPixels * numColors);
    int numFilters = hidActs.getNumRows() / numModules;

    int filterPixels = filterSize * filterSize;

    assert(hidActs.getNumCols() == numImages);

    assert(numModules * numFilters == hidActs.getNumRows());
    assert(numModulesX * numModulesX == numModules);
    assert(!images.isTrans());
    assert(!hidActs.isTrans());
    assert(hidActs.isContiguous());
    assert(numImages % 32 == 0);
    assert(numFilters % 16 == 0);

    targets.resize(numModules * numColors*filterPixels, numFilters);
    assert(!targets.isTrans());
    assert(targets.isContiguous());
    
    int preloadCases = 32;

    if (hidActsOrder == MODULE_FILTER_IMAGE) {
        if (numColors > 3) {
            // Worth playing with this parameter to find best value for your problem.
            // These values work relatively well, but not optimal for all problems.
            int pixelsPerThread = 2, colorsPerThread = numColors % 8 == 0 ? 8 : 4;
            int by = numFilters % 32 == 0 ? 4 : 8, bx = numFilters % 32 == 0 ? 32 : 16; // by == 4 seems to work best
            assert((by * bx) % preloadCases == 0);
            dim3 blocks(numModules*(numFilters/bx), DIVUP(filterPixels, by*pixelsPerThread) * (numColors / colorsPerThread));
            dim3 threads(bx, by);
            if (numColors % 8 == 0) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_manycolor<4,32,2,8,32,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_manycolor<4,32,2,8,32,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                                                   numImages, numFilters, numModulesX, imgSize, filterSize,
                                                                                   paddingStart, moduleStride, imgStride, numColors);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_manycolor<8,16,2,8,32,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_manycolor<8,16,2,8,32,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                                                   numImages, numFilters, numModulesX, imgSize, filterSize,
                                                                                   paddingStart, moduleStride, imgStride, numColors);
                }
            } else {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_manycolor<4,32,2,4,32,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_manycolor<4,32,2,4,32,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                                                   numImages, numFilters, numModulesX, imgSize, filterSize,
                                                                                   paddingStart, moduleStride, imgStride, numColors);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_manycolor<8,16,2,4,32,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_manycolor<8,16,2,4,32,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                                                   numImages, numFilters, numModulesX, imgSize, filterSize,
                                                                                   paddingStart, moduleStride, imgStride, numColors);
                }
            }

        } else {
            // Worth playing with this parameter to find best value for your problem.
            // These values work relatively well, but not optimal for all problems.
            int pixelsPerThread = numFilters % 32 == 0 ? (numColors == 1 ? 8 : 5) : (numColors == 1 ? 5 : 2);
            int by = numFilters % 32 == 0 ? 4 : 8, bx = numFilters % 32 == 0 ? 32 : 16; // by == 4 seems to work best
            assert((by * bx) % preloadCases == 0);
            dim3 blocks(numModules*(numFilters/bx), DIVUP(filterPixels, by*pixelsPerThread));
            dim3 threads(bx, by);
            // template <int B_Y, int B_X, int pixelsPerThread, int preloadCases, int colors>
            if (numColors == 1) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<4,32,8,32,1,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<4,32,8,32,1,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<8,16,5,32,1,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<8,16,5,32,1,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                }
            } else if (numColors == 2) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<4,32,5,32,2,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<4,32,5,32,2,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<8,16,2,32,2,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<8,16,2,32,2,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                }
            } else if (numColors == 3) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<4,32,5,32,3,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<4,32,5,32,3,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<8,16,2,32,3,true>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<8,16,2,32,3,true><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                }
            }
        }
    } else {
        if (numColors > 3) {
            // Worth playing with this parameter to find best value for your problem.
            // These values work relatively well, but not optimal for all problems.
            int pixelsPerThread = 2, colorsPerThread = numColors % 8 == 0 ? 8 : 4;
            int by = numFilters % 32 == 0 ? 4 : 8, bx = numFilters % 32 == 0 ? 32 : 16; // by == 4 seems to work best
            assert((by * bx) % preloadCases == 0);
            dim3 blocks(numModules*(numFilters/bx), DIVUP(filterPixels, by*pixelsPerThread) * (numColors / colorsPerThread));
            dim3 threads(bx, by);
            if (numColors % 8 == 0) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_manycolor<4,32,2,8,32,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_manycolor<4,32,2,8,32,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                                                   numImages, numFilters, numModulesX, imgSize, filterSize,
                                                                                   paddingStart, moduleStride, imgStride, numColors);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_manycolor<8,16,2,8,32,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_manycolor<8,16,2,8,32,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                                                   numImages, numFilters, numModulesX, imgSize, filterSize,
                                                                                   paddingStart, moduleStride, imgStride, numColors);
                }
            } else {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_manycolor<4,32,2,4,32,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_manycolor<4,32,2,4,32,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                                                   numImages, numFilters, numModulesX, imgSize, filterSize,
                                                                                   paddingStart, moduleStride, imgStride, numColors);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_manycolor<8,16,2,4,32,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_manycolor<8,16,2,4,32,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                                                   numImages, numFilters, numModulesX, imgSize, filterSize,
                                                                                   paddingStart, moduleStride, imgStride, numColors);
                }
            }

        } else {
            // Worth playing with this parameter to find best value for your problem.
            // These values work relatively well, but not optimal for all problems.
            int pixelsPerThread = numFilters % 32 == 0 ? (numColors == 1 ? 8 : 5) : (numColors == 1 ? 5 : 2);
            int by = numFilters % 32 == 0 ? 4 : 8, bx = numFilters % 32 == 0 ? 32 : 16; // by == 4 seems to work best
            assert((by * bx) % preloadCases == 0);
            dim3 blocks(numModules*(numFilters/bx), DIVUP(filterPixels, by*pixelsPerThread));
            dim3 threads(bx, by);
            // template <int B_Y, int B_X, int pixelsPerThread, int preloadCases, int colors>
            if (numColors == 1) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<4,32,8,32,1,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<4,32,8,32,1,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<8,16,5,32,1,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<8,16,5,32,1,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                }
            } else if (numColors == 2) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<4,32,5,32,2,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<4,32,5,32,2,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<8,16,2,32,2,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<8,16,2,32,2,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                }
            } else if (numColors == 3) {
                if (numFilters % 32 == 0) {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<4,32,5,32,3,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<4,32,5,32,3,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                } else {
                    cudaFuncSetCacheConfig(weight_acts_kernel_color<8,16,2,32,3,false>, cudaFuncCachePreferShared);
                    weight_acts_kernel_color<8,16,2,32,3,false><<<blocks, threads>>>(images.getDevData(), hidActs.getDevData(), targets.getDevData(),
                                                            numImages, numFilters, numModulesX, imgSize, filterSize, paddingStart, moduleStride, imgStride);
                }
            }
        }
    }

    cutilCheckMsg("computeWeightActs: kernel execution failed");
}
